import os
import pickle
import uuid

import dagstermill
from dagster import (
    Any,
    DependencyDefinition,
    Field,
    FileHandle,
    InputDefinition,
    Int,
    List,
    ModeDefinition,
    OutputDefinition,
    PipelineDefinition,
    ResourceDefinition,
    SolidDefinition,
    SolidInvocation,
    String,
    check,
    lambda_solid,
    pipeline,
    repository,
    resource,
    solid,
)
from dagster.utils import PICKLE_PROTOCOL

try:
    from dagster_pandas import DataFrame

    DAGSTER_PANDAS_PRESENT = True
except ImportError:
    DAGSTER_PANDAS_PRESENT = False

try:
    import sklearn as _

    SKLEARN_PRESENT = True
except ImportError:
    SKLEARN_PRESENT = False

try:
    import matplotlib as _

    MATPLOTLIB_PRESENT = True
except ImportError:
    MATPLOTLIB_PRESENT = False


def nb_test_path(name):
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "notebooks/{name}.ipynb".format(name=name)
    )


def define_hello_world_pipeline():
    return PipelineDefinition(name="hello_world_pipeline", solid_defs=[define_hello_world_solid()])


def define_hello_world_solid():
    return dagstermill.define_dagstermill_solid("hello_world", nb_test_path("hello_world"))


def define_hello_world_config_solid():
    return dagstermill.define_dagstermill_solid(
        "hello_world_config",
        nb_test_path("hello_world_config"),
        config_schema={"greeting": Field(String, is_required=False, default_value="hello")},
    )


def define_goodbye_config_solid():
    return dagstermill.define_dagstermill_solid(
        "goodbye_config",
        nb_test_path("print_dagstermill_context_solid_config"),
        config_schema={"farewell": Field(String, is_required=False, default_value="goodbye")},
    )


def define_hello_world_config_pipeline():
    return PipelineDefinition(
        name="hello_world_config_pipeline",
        solid_defs=[define_hello_world_config_solid(), define_goodbye_config_solid()],
    )


def define_hello_world_with_output_notebook_solid():
    return dagstermill.define_dagstermill_solid(
        "hello_world_with_output_notebook", nb_test_path("hello_world"), output_notebook="notebook",
    )


def define_hello_world_with_output_notebook_pipeline():
    @solid(input_defs=[InputDefinition("notebook", dagster_type=FileHandle)])
    def load_notebook_solid(_, notebook):
        return os.path.exists(notebook.path_desc)

    return PipelineDefinition(
        name="hello_world_with_output_notebook_pipeline",
        solid_defs=[define_hello_world_with_output_notebook_solid(), load_notebook_solid],
        dependencies={
            "load_notebook_solid": {
                "notebook": DependencyDefinition("hello_world_with_output_notebook", "notebook")
            }
        },
    )


def define_hello_world_with_output():
    return dagstermill.define_dagstermill_solid(
        "hello_world_output", nb_test_path("hello_world_output"), [], [OutputDefinition()]
    )


def define_hello_world_with_output_pipeline():
    return PipelineDefinition(
        name="hello_world_with_output_pipeline", solid_defs=[define_hello_world_with_output()]
    )


def define_hello_world_explicit_yield():
    return dagstermill.define_dagstermill_solid(
        "hello_world_explicit_yield_pipeline",
        nb_test_path("hello_world_explicit_yield"),
        [],
        [OutputDefinition()],
    )


def define_hello_world_explicit_yield_pipeline():
    return PipelineDefinition(
        name="hello_world_explicit_yield_pipeline", solid_defs=[define_hello_world_explicit_yield()]
    )


def define_hello_logging_solid():
    return dagstermill.define_dagstermill_solid("hello_logging", nb_test_path("hello_logging"))


def define_hello_logging_pipeline():
    return PipelineDefinition(
        name="hello_logging_pipeline", solid_defs=[define_hello_logging_solid()]
    )


# This probably should be moved to a library because it is immensely useful for testing
def solid_definition(fn):
    return check.inst(fn(), SolidDefinition)


@solid_definition
def add_two_numbers_pm_solid():
    return dagstermill.define_dagstermill_solid(
        "add_two_numbers",
        nb_test_path("add_two_numbers"),
        [InputDefinition(name="a", dagster_type=Int), InputDefinition(name="b", dagster_type=Int)],
        [OutputDefinition(Int)],
    )


@solid_definition
def mult_two_numbers_pm_solid():
    return dagstermill.define_dagstermill_solid(
        "mult_two_numbers",
        nb_test_path("mult_two_numbers"),
        [InputDefinition(name="a", dagster_type=Int), InputDefinition(name="b", dagster_type=Int)],
        [OutputDefinition(Int)],
    )


@lambda_solid
def return_one():
    return 1


@lambda_solid
def return_two():
    return 2


def define_add_pipeline():
    add_two_numbers = add_two_numbers_pm_solid
    return PipelineDefinition(
        name="test_add_pipeline",
        solid_defs=[return_one, return_two, add_two_numbers],
        dependencies={
            add_two_numbers.name: {
                "a": DependencyDefinition("return_one"),
                "b": DependencyDefinition("return_two"),
            }
        },
    )


@solid(input_defs=[], config_schema=Int)
def load_constant(context):
    return context.solid_config


def define_test_notebook_dag_pipeline():
    return PipelineDefinition(
        name="test_notebook_dag",
        solid_defs=[load_constant, add_two_numbers_pm_solid, mult_two_numbers_pm_solid],
        dependencies={
            SolidInvocation("load_constant", alias="load_a"): {},
            SolidInvocation("load_constant", alias="load_b"): {},
            SolidInvocation(name="add_two_numbers", alias="add_two"): {
                "a": DependencyDefinition("load_a"),
                "b": DependencyDefinition("load_b"),
            },
            SolidInvocation(name="mult_two_numbers", alias="mult_two"): {
                "a": DependencyDefinition("add_two"),
                "b": DependencyDefinition("load_b"),
            },
        },
    )


def define_error_pipeline():
    return PipelineDefinition(
        name="error_pipeline",
        solid_defs=[
            dagstermill.define_dagstermill_solid("error_solid", nb_test_path("error_notebook"))
        ],
    )


if DAGSTER_PANDAS_PRESENT and SKLEARN_PRESENT and MATPLOTLIB_PRESENT:

    @solid_definition
    def clean_data_solid():
        return dagstermill.define_dagstermill_solid(
            "clean_data", nb_test_path("clean_data"), output_defs=[OutputDefinition(DataFrame)]
        )

    @solid_definition
    def LR_solid():
        return dagstermill.define_dagstermill_solid(
            "linear_regression",
            nb_test_path("tutorial_LR"),
            input_defs=[InputDefinition(name="df", dagster_type=DataFrame)],
        )

    @solid_definition
    def RF_solid():
        return dagstermill.define_dagstermill_solid(
            "random_forest_regression",
            nb_test_path("tutorial_RF"),
            input_defs=[InputDefinition(name="df", dagster_type=DataFrame)],
        )

    def define_tutorial_pipeline():
        return PipelineDefinition(
            name="tutorial_pipeline",
            solid_defs=[clean_data_solid, LR_solid, RF_solid],
            dependencies={
                SolidInvocation("clean_data"): {},
                SolidInvocation("linear_regression"): {"df": DependencyDefinition("clean_data")},
                SolidInvocation("random_forest_regression"): {
                    "df": DependencyDefinition("clean_data")
                },
            },
        )


@solid("resource_solid", required_resource_keys={"list"})
def resource_solid(context):
    context.resources.list.append("Hello, solid!")
    return True


@solid_definition
def hello_world_resource_solid():
    return dagstermill.define_dagstermill_solid(
        "hello_world_resource",
        nb_test_path("hello_world_resource"),
        input_defs=[InputDefinition("nonce")],
        required_resource_keys={"list"},
    )


@solid_definition
def hello_world_resource_with_exception_solid():
    return dagstermill.define_dagstermill_solid(
        "hello_world_resource_with_exception",
        nb_test_path("hello_world_resource_with_exception"),
        input_defs=[InputDefinition("nonce")],
        required_resource_keys={"list"},
    )


class FilePickleList(object):
    # This is not thread- or anything else-safe
    def __init__(self, path):
        self.closed = False
        self.id = str(uuid.uuid4())[-6:]
        self.path = path
        self.list = []
        if not os.path.exists(self.path):
            self.write()
        self.read()
        self.open()

    def open(self):
        self.read()
        self.append("Opened")

    def append(self, obj):
        self.read()
        self.list.append(self.id + ": " + obj)
        self.write()

    def read(self):
        with open(self.path, "rb") as fd:
            self.list = pickle.load(fd)
            return self.list

    def write(self):
        with open(self.path, "wb") as fd:
            pickle.dump(self.list, fd, protocol=PICKLE_PROTOCOL)

    def close(self):
        self.append("Closed")
        self.closed = True


@resource(config_schema=Field(String))
def filepicklelist_resource(init_context):
    filepicklelist = FilePickleList(init_context.resource_config)
    try:
        yield filepicklelist
    finally:
        filepicklelist.close()


def define_resource_pipeline():
    return PipelineDefinition(
        name="resource_pipeline",
        solid_defs=[resource_solid, hello_world_resource_solid],
        dependencies={"hello_world_resource": {"nonce": DependencyDefinition("resource_solid")}},
        mode_defs=[
            ModeDefinition(name="test", resource_defs={"list": ResourceDefinition(lambda _: [])}),
            ModeDefinition(name="prod", resource_defs={"list": filepicklelist_resource}),
        ],
    )


def define_resource_with_exception_pipeline():
    return PipelineDefinition(
        name="resource_with_exception_pipeline",
        solid_defs=[resource_solid, hello_world_resource_with_exception_solid],
        dependencies={
            "hello_world_resource_with_exception": {"nonce": DependencyDefinition("resource_solid")}
        },
        mode_defs=[ModeDefinition(resource_defs={"list": filepicklelist_resource})],
    )


@solid_definition
def bad_kernel_solid():
    return dagstermill.define_dagstermill_solid("bad_kernel_solid", nb_test_path("bad_kernel"))


def define_bad_kernel_pipeline():
    return PipelineDefinition(name="bad_kernel_pipeline", solid_defs=[bad_kernel_solid])


reimport_solid = dagstermill.define_dagstermill_solid(
    "reimport",
    nb_test_path("reimport"),
    input_defs=[InputDefinition("l", List[int])],
    output_defs=[OutputDefinition()],
)


@solid
def lister(_):
    return [1, 2, 3]


@pipeline
def reimport_pipeline():
    reimport_solid(lister())


@solid_definition
def yield_3_solid():
    return dagstermill.define_dagstermill_solid(
        "yield_3", nb_test_path("yield_3"), [], [OutputDefinition(Int)],
    )


@pipeline
def yield_3_pipeline():
    yield_3_solid()


class BasicTest(object):
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return "BasicTest: {x}".format(x=str(self.x))


@solid_definition
def yield_obj_solid():
    return dagstermill.define_dagstermill_solid(
        "yield_obj", nb_test_path("yield_obj"), [], [OutputDefinition(Any)],
    )


@pipeline
def yield_obj_pipeline():
    yield_obj_solid()


@repository
def notebook_repo():
    pipeline_dict = {
        "bad_kernel_pipeline": define_bad_kernel_pipeline,
        "error_pipeline": define_error_pipeline,
        "hello_world_pipeline": define_hello_world_pipeline,
        "hello_world_config_pipeline": define_hello_world_config_pipeline,
        "hello_world_explicit_yield_pipeline": define_hello_world_explicit_yield_pipeline,
        "hello_world_with_output_pipeline": define_hello_world_with_output_pipeline,
        "hello_logging_pipeline": define_hello_logging_pipeline,
        "resource_pipeline": define_resource_pipeline,
        "resource_with_exception_pipeline": define_resource_with_exception_pipeline,
        "test_add_pipeline": define_add_pipeline,
        "test_notebook_dag": define_test_notebook_dag_pipeline,
        "reimport_pipeline": lambda: reimport_pipeline,
        "yield_3_pipeline": lambda: yield_3_pipeline,
        "yield_obj_pipeline": lambda: yield_obj_pipeline,
    }
    if DAGSTER_PANDAS_PRESENT and SKLEARN_PRESENT and MATPLOTLIB_PRESENT:
        pipeline_dict["tutorial_pipeline"] = define_tutorial_pipeline

    return {"pipelines": pipeline_dict}



import nbformat
from nbconvert.preprocessors.execute import CellExecutionError
from papermill.engines import NBConvertEngine
from papermill.log import logger
from papermill.preprocess import PapermillExecutePreprocessor


class DagstermillExecutePreprocessor(PapermillExecutePreprocessor):
    # We need to finalize dagster resources here (as opposed to, e.g., in the notebook_complete
    # method on the NotebookExecutionManager), because we need to be inside the scope of the
    # nbconvert.preprocessors.ExecutePreprocessor.setup_preprocessor context manager, which tears
    # the kernel down. Note that atexit doesn't seem to work at all in ipython, and hooking into
    # the ipython post_execute event doesn't work in papermill.
    def papermill_process(self, nb_man, resources):
        _, resources = super(DagstermillExecutePreprocessor, self).papermill_process(
            nb_man, resources
        )

        new_cell = nbformat.v4.new_code_cell(
            source=("import dagstermill as __dm_dagstermill\n" "__dm_dagstermill._teardown()\n")
        )
        new_cell.metadata["tags"] = ["injected-teardown"]
        new_cell.metadata["papermill"] = {}
        index = len(nb_man.nb.cells)
        nb_man.nb.cells = nb_man.nb.cells + [new_cell]

        # Calqued from PapermillExecutePreprocessor.papermill_process
        try:
            nb_man.cell_start(new_cell, index)
            nb_man.nb.cells[index], _ = self.preprocess_cell(new_cell, None, index)
        except CellExecutionError as ex:  # pragma: nocover
            nb_man.cell_exception(nb_man.nb.cells[index], cell_index=index, exception=ex)
        finally:
            nb_man.cell_complete(nb_man.nb.cells[index], cell_index=index)

        return nb_man.nb, resources


class DagstermillNBConvertEngine(NBConvertEngine):
    @classmethod
    def execute_managed_notebook(
        cls,
        nb_man,
        kernel_name,
        log_output=False,
        stdout_file=None,
        stderr_file=None,
        start_timeout=60,
        execution_timeout=None,
        **kwargs
    ):
        # Nicely handle preprocessor arguments prioritizing values set by engine
        preprocessor = DagstermillExecutePreprocessor(
            timeout=execution_timeout if execution_timeout else kwargs.get("timeout"),
            startup_timeout=start_timeout,
            kernel_name=kernel_name,
            log=logger,
        )

        preprocessor.log_output = log_output  # pylint:disable = attribute-defined-outside-init
        preprocessor.preprocess(nb_man, kwargs)



from dagster.core.utils import check_dagster_package_version

from .context import DagstermillExecutionContext
from .errors import DagstermillError, DagstermillExecutionError
from .manager import MANAGER_FOR_NOTEBOOK_INSTANCE as _MANAGER_FOR_NOTEBOOK_INSTANCE
from .solids import define_dagstermill_solid
from .version import __version__

check_dagster_package_version("dagstermill", __version__)

get_context = _MANAGER_FOR_NOTEBOOK_INSTANCE.get_context

yield_result = _MANAGER_FOR_NOTEBOOK_INSTANCE.yield_result

yield_event = _MANAGER_FOR_NOTEBOOK_INSTANCE.yield_event

_reconstitute_pipeline_context = _MANAGER_FOR_NOTEBOOK_INSTANCE.reconstitute_pipeline_context

_teardown = _MANAGER_FOR_NOTEBOOK_INSTANCE.teardown_resources

_load_parameter = _MANAGER_FOR_NOTEBOOK_INSTANCE.load_parameter



from setuptools import find_packages, setup


def get_version():
    version = {}
    with open("dagstermill/version.py") as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    return version["__version__"]


if __name__ == "__main__":
    setup(
        name="dagstermill",
        version=get_version(),
        author="Elementl",
        author_email="hello@elementl.com",
        license="Apache-2.0",
        packages=find_packages(exclude=["dagstermill_tests"]),
        classifiers=[
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        install_requires=[
            "dagster",
            "ipykernel>=4.9.0",
            "nbconvert>=5.4.0,<6.0.0",
            "nteract-scrapbook>=0.2.0",
            "papermill>=1.0.0,<2.0.0",
            "six",
        ],
        entry_points={"console_scripts": ["dagstermill = dagstermill.cli:main"]},
    )



import copy
import os
import pickle
import uuid

import nbformat
import papermill
import six
from dagster import (
    AssetMaterialization,
    EventMetadataEntry,
    InputDefinition,
    Output,
    OutputDefinition,
    SolidDefinition,
    check,
    seven,
)
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.core.definitions.reconstructable import ReconstructablePipeline
from dagster.core.errors import user_code_error_boundary
from dagster.core.execution.context.compute import SolidExecutionContext
from dagster.core.execution.context.system import SystemComputeExecutionContext
from dagster.core.storage.file_manager import FileHandle
from dagster.serdes import pack_value
from dagster.utils import mkdir_p, safe_tempfile_path
from papermill.engines import papermill_engines
from papermill.iorw import load_notebook_node, write_ipynb
from papermill.parameterize import _find_first_tagged_cell_index

from .engine import DagstermillNBConvertEngine
from .errors import DagstermillError, DagstermillExecutionError
from .serialize import read_value, write_value
from .translator import RESERVED_INPUT_NAMES, DagsterTranslator


# This is based on papermill.parameterize.parameterize_notebook
# Typically, papermill injects the injected-parameters cell *below* the parameters cell
# but we want to *replace* the parameters cell, which is what this function does.
def replace_parameters(context, nb, parameters):
    """Assigned parameters into the appropiate place in the input notebook

    Args:
        nb (NotebookNode): Executable notebook object
        parameters (dict): Arbitrary keyword arguments to pass to the notebook parameters.
    """
    check.dict_param(parameters, "parameters")

    # Copy the nb object to avoid polluting the input
    nb = copy.deepcopy(nb)

    # papermill method chooses translator based on kernel_name and language, but we just call the
    # DagsterTranslator to generate parameter content based on the kernel_name
    param_content = DagsterTranslator.codify(parameters)

    newcell = nbformat.v4.new_code_cell(source=param_content)
    newcell.metadata["tags"] = ["injected-parameters"]

    param_cell_index = _find_first_tagged_cell_index(nb, "parameters")
    injected_cell_index = _find_first_tagged_cell_index(nb, "injected-parameters")
    if injected_cell_index >= 0:
        # Replace the injected cell with a new version
        before = nb.cells[:injected_cell_index]
        after = nb.cells[injected_cell_index + 1 :]
        check.int_value_param(param_cell_index, -1, "param_cell_index")
        # We should have blown away the parameters cell if there is an injected-parameters cell
    elif param_cell_index >= 0:
        # Replace the parameter cell with the injected-parameters cell
        before = nb.cells[:param_cell_index]
        after = nb.cells[param_cell_index + 1 :]
    else:
        # Inject to the top of the notebook, presumably first cell includes dagstermill import
        context.log.debug(
            (
                "Executing notebook with no tagged parameters cell: injecting boilerplate in first "
                "cell."
            )
        )
        before = []
        after = nb.cells

    nb.cells = before + [newcell] + after
    nb.metadata.papermill["parameters"] = seven.json.dumps(parameters)

    return nb


def get_papermill_parameters(compute_context, inputs, output_log_path):
    check.inst_param(compute_context, "compute_context", SystemComputeExecutionContext)
    check.param_invariant(
        isinstance(compute_context.run_config, dict),
        "compute_context",
        "SystemComputeExecutionContext must have valid run_config",
    )
    check.dict_param(inputs, "inputs", key_type=six.string_types)

    run_id = compute_context.run_id

    marshal_dir = "/tmp/dagstermill/{run_id}/marshal".format(run_id=run_id)
    mkdir_p(marshal_dir)

    if not isinstance(compute_context.pipeline, ReconstructablePipeline):
        raise DagstermillError(
            "Can't execute a dagstermill solid from a pipeline that is not reconstructable. "
            "Use the reconstructable() function if executing from python"
        )

    dm_executable_dict = compute_context.pipeline.to_dict()

    dm_context_dict = {
        "output_log_path": output_log_path,
        "marshal_dir": marshal_dir,
        "run_config": compute_context.run_config,
    }

    dm_solid_handle_kwargs = compute_context.solid_handle._asdict()

    parameters = {}

    input_def_dict = compute_context.solid_def.input_dict
    for input_name, input_value in inputs.items():
        assert (
            input_name not in RESERVED_INPUT_NAMES
        ), "Dagstermill solids cannot have inputs named {input_name}".format(input_name=input_name)
        dagster_type = input_def_dict[input_name].dagster_type
        parameter_value = write_value(
            dagster_type, input_value, os.path.join(marshal_dir, "input-{}".format(input_name))
        )
        parameters[input_name] = parameter_value

    parameters["__dm_context"] = dm_context_dict
    parameters["__dm_executable_dict"] = dm_executable_dict
    parameters["__dm_pipeline_run_dict"] = pack_value(compute_context.pipeline_run)
    parameters["__dm_solid_handle_kwargs"] = dm_solid_handle_kwargs
    parameters["__dm_instance_ref_dict"] = pack_value(compute_context.instance.get_ref())

    return parameters


def _dm_solid_compute(name, notebook_path, output_notebook=None):
    check.str_param(name, "name")
    check.str_param(notebook_path, "notebook_path")
    check.opt_str_param(output_notebook, "output_notebook")

    def _t_fn(compute_context, inputs):
        check.inst_param(compute_context, "compute_context", SolidExecutionContext)
        check.param_invariant(
            isinstance(compute_context.run_config, dict),
            "context",
            "SystemComputeExecutionContext must have valid run_config",
        )

        system_compute_context = compute_context.get_system_context()

        with seven.TemporaryDirectory() as output_notebook_dir:
            with safe_tempfile_path() as output_log_path:

                parameterized_notebook_path = os.path.join(
                    output_notebook_dir, "{prefix}-inter.ipynb".format(prefix=str(uuid.uuid4()))
                )

                executed_notebook_path = os.path.join(
                    output_notebook_dir, "{prefix}-out.ipynb".format(prefix=str(uuid.uuid4()))
                )

                # Scaffold the registration here
                nb = load_notebook_node(notebook_path)
                nb_no_parameters = replace_parameters(
                    system_compute_context,
                    nb,
                    get_papermill_parameters(system_compute_context, inputs, output_log_path),
                )
                write_ipynb(nb_no_parameters, parameterized_notebook_path)

                with user_code_error_boundary(
                    DagstermillExecutionError,
                    lambda: (
                        "Error occurred during the execution of Dagstermill solid "
                        "{solid_name}: {notebook_path}".format(
                            solid_name=name, notebook_path=notebook_path
                        )
                    ),
                ):
                    try:
                        papermill_engines.register("dagstermill", DagstermillNBConvertEngine)
                        papermill.execute_notebook(
                            input_path=parameterized_notebook_path,
                            output_path=executed_notebook_path,
                            engine_name="dagstermill",
                            log_output=True,
                        )

                    except Exception as exc:  # pylint: disable=broad-except
                        try:
                            with open(executed_notebook_path, "rb") as fd:
                                executed_notebook_file_handle = compute_context.file_manager.write(
                                    fd, mode="wb", ext="ipynb"
                                )
                                executed_notebook_materialization_path = (
                                    executed_notebook_file_handle.path_desc
                                )
                        except Exception as exc_inner:  # pylint: disable=broad-except
                            compute_context.log.warning(
                                "Error when attempting to materialize executed notebook using file manager (falling back to local): {exc}".format(
                                    exc=exc_inner
                                )
                            )
                            executed_notebook_materialization_path = executed_notebook_path

                        yield AssetMaterialization(
                            asset_key=executed_notebook_materialization_path,
                            description="Location of output notebook in file manager",
                            metadata_entries=[
                                EventMetadataEntry.fspath(executed_notebook_materialization_path)
                            ],
                        )
                        raise exc

            system_compute_context.log.debug(
                "Notebook execution complete for {name} at {executed_notebook_path}.".format(
                    name=name, executed_notebook_path=executed_notebook_path,
                )
            )

            executed_notebook_file_handle = None
            try:
                # use binary mode when when moving the file since certain file_managers such as S3
                # may try to hash the contents
                with open(executed_notebook_path, "rb") as fd:
                    executed_notebook_file_handle = compute_context.file_manager.write(
                        fd, mode="wb", ext="ipynb"
                    )
                    executed_notebook_materialization_path = executed_notebook_file_handle.path_desc
            except Exception as exc:  # pylint: disable=broad-except
                compute_context.log.warning(
                    "Error when attempting to materialize executed notebook using file manager (falling back to local): {exc}".format(
                        exc=str(exc)
                    )
                )
                executed_notebook_materialization_path = executed_notebook_path

            yield AssetMaterialization(
                asset_key=executed_notebook_materialization_path,
                description="Location of output notebook in file manager",
                metadata_entries=[
                    EventMetadataEntry.fspath(executed_notebook_materialization_path)
                ],
            )

            if output_notebook is not None:
                yield Output(executed_notebook_file_handle, output_notebook)

            # deferred import for perf
            import scrapbook

            output_nb = scrapbook.read_notebook(executed_notebook_path)

            for (output_name, output_def) in system_compute_context.solid_def.output_dict.items():
                data_dict = output_nb.scraps.data_dict
                if output_name in data_dict:
                    value = read_value(output_def.dagster_type, data_dict[output_name])

                    yield Output(value, output_name)

            for key, value in output_nb.scraps.items():
                if key.startswith("event-"):
                    with open(value.data, "rb") as fd:
                        yield pickle.loads(fd.read())

    return _t_fn


def define_dagstermill_solid(
    name,
    notebook_path,
    input_defs=None,
    output_defs=None,
    config_schema=None,
    required_resource_keys=None,
    output_notebook=None,
):
    """Wrap a Jupyter notebook in a solid.

    Arguments:
        name (str): The name of the solid.
        notebook_path (str): Path to the backing notebook.
        input_defs (Optional[List[InputDefinition]]): The solid's inputs.
        output_defs (Optional[List[OutputDefinition]]): The solid's outputs. Your notebook should
            call :py:func:`~dagstermill.yield_result` to yield each of these outputs.
        required_resource_keys (Optional[Set[str]]): The string names of any required resources.
        output_notebook (Optional[str]): If set, will be used as the name of an injected output of
            type :py:class:`~dagster.FileHandle` that will point to the executed notebook (in
            addition to the :py:class:`~dagster.AssetMaterialization` that is always created). This
            respects the :py:class:`~dagster.core.storage.file_manager.FileManager` configured on
            the pipeline system storage, so, e.g., if :py:class:`~dagster_aws.s3.s3_system_storage`
            is configured, the output will be a :py:class:`~dagster_aws.s3.S3FileHandle`.

    Returns:
        :py:class:`~dagster.SolidDefinition`
    """
    check.str_param(name, "name")
    check.str_param(notebook_path, "notebook_path")
    input_defs = check.opt_list_param(input_defs, "input_defs", of_type=InputDefinition)
    output_defs = check.opt_list_param(output_defs, "output_defs", of_type=OutputDefinition)
    required_resource_keys = check.opt_set_param(
        required_resource_keys, "required_resource_keys", of_type=str
    )

    return SolidDefinition(
        name=name,
        input_defs=input_defs,
        compute_fn=_dm_solid_compute(name, notebook_path, output_notebook),
        output_defs=output_defs
        + (
            [OutputDefinition(dagster_type=FileHandle, name=output_notebook)]
            if output_notebook
            else []
        ),
        config_schema=check_user_facing_opt_config_param(config_schema, "config_schema"),
        required_resource_keys=required_resource_keys,
        description="This solid is backed by the notebook at {path}".format(path=notebook_path),
        tags={"notebook_path": notebook_path, "kind": "ipynb"},
    )



from dagster.core.errors import DagsterError, DagsterUserCodeExecutionError


class DagstermillError(DagsterError):
    """Base class for errors raised by dagstermill."""


class DagstermillExecutionError(DagstermillError, DagsterUserCodeExecutionError):
    """Errors raised during the execution of dagstermill solids."""



from dagster import check, seven
from dagster.core.types.dagster_type import DagsterType, DagsterTypeKind

PICKLE_PROTOCOL = 2


def is_json_serializable(value):
    try:
        seven.json.dumps(value)
        return True
    except TypeError:
        return False


def read_value(dagster_type, value):
    check.inst_param(dagster_type, "dagster_type", DagsterType)
    check.dict_param(value, "value")
    check.invariant(
        list(value.keys()) in [["value"], ["file"]],
        "Malformed value received with keys: {bad_keys}, expected ['value'] or ['file']".format(
            bad_keys="[{keys}]".format(
                keys=", ".join(["'{key}'".format(key=key) for key in value.keys()])
            )
        ),
    )
    if "value" in value:
        return value["value"]
    else:
        return dagster_type.serialization_strategy.deserialize_from_file(value["file"])


def write_value(dagster_type, value, target_file):
    check.inst_param(dagster_type, "dagster_type", DagsterType)
    if dagster_type.kind == DagsterTypeKind.SCALAR or (
        dagster_type.kind == DagsterTypeKind.ANY and is_json_serializable(value)
    ):
        return {"value": value}
    else:
        dagster_type.serialization_strategy.serialize_to_file(value, target_file)
        return {"file": target_file}



import copy
import os
import subprocess

import click
import nbformat
from dagster import check
from dagster.seven.json import loads
from dagster.utils import mkdir_p, safe_isfile
from papermill.iorw import load_notebook_node, write_ipynb


def get_import_cell():
    return nbformat.v4.new_code_cell(source="import dagstermill")


def get_parameters_cell():
    parameters_cell = nbformat.v4.new_code_cell(source="context = dagstermill.get_context()")
    parameters_cell.metadata["tags"] = ["parameters"]
    return parameters_cell


def get_kernelspec(kernel):
    kernelspecs = loads(subprocess.check_output(["jupyter", "kernelspec", "list", "--json"]))

    check.invariant(len(kernelspecs["kernelspecs"]) > 0, "No available Jupyter kernelspecs!")

    if kernel is None:
        preferred_kernels = list(
            filter(
                lambda kernel_name: kernel_name in kernelspecs["kernelspecs"],
                ["dagster", "python3", "python"],
            )
        ) + list(kernelspecs["kernelspecs"].keys())
        kernel = preferred_kernels[0]
        print(  # pylint: disable=print-call
            "No kernel specified, defaulting to '{kernel}'".format(kernel=kernel)
        )

    check.invariant(
        kernel in kernelspecs["kernelspecs"],
        "Could not find kernel '{kernel}': available kernels are [{kernels}]".format(
            kernel=kernel,
            kernels=", ".join(["'{k}'".format(k=k) for k in kernelspecs["kernelspecs"]]),
        ),
    )

    return {
        "name": kernel,
        "language": kernelspecs["kernelspecs"][kernel]["spec"]["language"],
        "display_name": kernelspecs["kernelspecs"][kernel]["spec"]["display_name"],
    }


def get_notebook_scaffolding(kernelspec):
    check.dict_param(kernelspec, "kernelspec", key_type=str, value_type=str)

    notebook = nbformat.v4.new_notebook()

    notebook.cells = [get_import_cell(), get_parameters_cell()]

    metadata = {"celltoolbar": "Tags", "kernelspec": kernelspec}

    notebook.metadata = metadata

    return nbformat.writes(notebook)


@click.command(
    name="register-notebook", help=("Scaffolds existing notebook for dagstermill compatibility")
)
@click.option("--notebook", "-note", type=click.Path(exists=True), help="Path to existing notebook")
def retroactively_scaffold_notebook(notebook):
    execute_retroactive_scaffold(notebook)


def execute_retroactive_scaffold(notebook_path):
    nb = load_notebook_node(notebook_path)
    new_nb = copy.deepcopy(nb)
    new_nb.cells = [get_import_cell(), get_parameters_cell()] + nb.cells
    write_ipynb(new_nb, notebook_path)


@click.command(name="create-notebook", help=("Creates new dagstermill notebook."))
@click.option("--notebook", "-note", type=click.Path(), help="Name of notebook")
@click.option(
    "--force-overwrite",
    is_flag=True,
    help="Will force overwrite any existing notebook or file with the same name.",
)
@click.option(
    "--kernel",
    type=click.STRING,
    help=(
        "Specify an existing Jupyter kernel to use. (Run `jupyter kernelspec list` to view "
        "available kernels.) By default a kernel called 'dagster' will be used, if available, "
        "then one called 'python3', then 'python', and then the first available kernel."
    ),
)
def create_notebook(notebook, force_overwrite, kernel):
    execute_create_notebook(notebook, force_overwrite, kernel)


def execute_create_notebook(notebook, force_overwrite, kernel):
    notebook_path = os.path.join(
        os.getcwd(), notebook if notebook.endswith(".ipynb") else notebook + ".ipynb"
    )

    notebook_dir = os.path.dirname(notebook_path)
    mkdir_p(notebook_dir)

    if not force_overwrite and safe_isfile(notebook_path):
        click.confirm(
            (
                "Warning, {notebook_path} already exists and continuing "
                "will overwrite the existing notebook. "
                "Are you sure you want to continue?"
            ).format(notebook_path=notebook_path),
            abort=True,
        )

    with open(notebook_path, "w") as f:
        f.write(get_notebook_scaffolding(get_kernelspec(kernel)))
        click.echo("Created new dagstermill notebook at {path}".format(path=notebook_path))


def create_dagstermill_cli():
    group = click.Group(name="dagstermill")
    group.add_command(create_notebook)
    group.add_command(retroactively_scaffold_notebook)
    return group


def main():
    cli = create_dagstermill_cli()
    # click magic
    cli(obj={})  # pylint:disable=E1120
